#include <sys/types.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <sys/wait.h>
#include <sys/stat.h>
#include <sys/signal.h>
#include <sys/time.h>
#include <sys/param.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <stdarg.h>

#include <netinet/in.h>
#include <arpa/inet.h>

#include <netdb.h>
#include <stdio.h>
#include <errno.h>
#include <ctype.h>
#include <syslog.h>
#include <string.h>

#include "types.h"
#include "config.h"
#include "globals.h"
#include "logfun.h"
#include "fileio.h"
#include "process.h"
#include "readconfig.h"
#include "netio.h"

#ifdef MAX
#undef MAX
#endif
#define MAX(A, B)	(((A) >= (B)) ? (A) : (B))

static CHAR* tftpErrors[] =  {	NULL,
				"File not found",
				"Access violation",
				"Disk full or allocation exceeded",
				"Illegal TFTP operation",
				"Unknown transfer ID",
				"File already exists",
				"No such user"		};

/*
	basically same as the library function inet_ntoa
	but tries to resolve the host address into a host name
	if not prohibited by 'noResolv'
*/

CHAR *xinet_ntoa (in_addr_t ia)
{
    struct hostent *he;
    struct in_addr sia;

    if (!noResolv) {
        he = gethostbyaddr ((CHAR *) &ia, sizeof (in_addr_t), AF_INET);
        if (he != NULL)
            return he->h_name;
    }

    sia.s_addr = ia;
    return inet_ntoa(sia);
}

/*
	check if this file descriptor refers to a socket
	used to determine if we're forked by inetd or started standalone
*/

INT isaSocketFd (INT fd)
{
    INT dummy, length = sizeof(INT);

    return (!getsockopt(fd, SOL_SOCKET, SO_TYPE, &dummy, &length));
}

/*
	check if this file pointer refers to a socket
	used to determine if we're forked by inetd or started standalone
*/

INT isaSocket (FILE *fp)
{
    return (isaSocketFd(fileno(fp)));
}

/*
	send an error message
*/

INT sendToError (sopair_t *sockets, sai_t *peer, INT err, ...)
{
    va_list ap;
    tftpD_t errBuffer;
    struct timeval sendTimeout, recvTimeout;
    fd_set fdSel;
    INT ns, ios;
    CHAR *errText;

    errBuffer.head.opCode  = htons(RT_ERROR);
    errBuffer.head.errCode = htons(err);

    switch (err) {
        case PE_UNDEF:
            errText = va_arg(ap, CHAR *);
            va_start (ap, err);
            strncpy(errBuffer.data, errText, SEGSIZE);
            break;
        case PE_NOTFOUND:
        case PE_ACCESS:
        case PE_NOSPACE:
        case PE_BADOP:
        case PE_BADID:
        case PE_EXISTS:
        case PE_NOUSER:
            strncpy(errBuffer.data, tftpErrors[err], SEGSIZE);
            break;
        default:
            return -1;
    }

    LOG_MSG4 (LOG_DEBUG, "Sending error message: \"%s\"",errBuffer.data);

    ns = 1 + MAX(sockets->recv, sockets->send);

    errBuffer.data[SEGSIZE-1] = 0;

    sendTimeout.tv_sec  = SEND_TIMEOUT;
    sendTimeout.tv_usec = 0;
    FD_ZERO(&fdSel);
    FD_SET(sockets->send, &fdSel);

    if (select (ns, NULL, &fdSel, NULL, &sendTimeout) <= 0) {
        LOG_MSG1 (LOG_ERR, "Error during select (write/ERR): %m");
        return -1;
    }

    if (sendto(sockets->send, &errBuffer,
               sizeof(tftpH_t)+strlen(errBuffer.data)+1, 0,
               (sa_t *)peer, sizeof(sai_t)) < 0) {
        LOG_MSG1 (LOG_ERR, "Cannot send error message: %m");
        return -1;
    }

/*
	the protocol allows for ACK's on error packets.
	we are prepared by waiting some time for a packet like that
*/
    recvTimeout.tv_sec  = RECV_TIMEOUT;
    recvTimeout.tv_usec = 0;
    FD_ZERO(&fdSel);
    FD_SET(sockets->recv, &fdSel);

    if ((ios = select (ns, &fdSel, NULL, NULL, &recvTimeout)) < 0) {
        LOG_MSG1 (LOG_ERR, "Error during select (read/ERR): %m");
        return -1;
    }

    if (ios) {
	tftpH_t ackBuffer;
        INT recvLength;

        recvLength = sizeof(tftpH_t);
        recvLength = recv (sockets->recv, &ackBuffer, recvLength, 0);
        if (recvLength < 0) {
            LOG_MSG1 (LOG_ERR, "Error during recv/ERR: %m");
            return -1;
        }

        if (recvLength < (INT)sizeof(tftpH_t)) {
            LOG_MSG1 (LOG_ERR, "Short reply packet after ERR");
            return -1;
        }

        if (ntohs(ackBuffer.opCode) != RT_ACK) {
            LOG_MSG1 (LOG_ERR, "Non ACK packet reply on ERR");
            return -1;
        }
        else {
            LOG_MSG4 (LOG_DEBUG, "Peer sent ACK on ERR");
        }
    }

    return 0;
}

/*
	send an error message to the current peer
*/

INT sendError (tftpR_t *request, INT err, ...)
{
    va_list ap;
    sopair_t sockets;

    sockets.recv = request->Rsock;
    sockets.send = request->Ssock;

    if (err == PE_UNDEF) {
        CHAR *errText = va_arg(ap, CHAR *);
        va_start (ap, err);
        return (sendToError(&sockets, &request->reqPeer, err, errText));
    }
    else
        return (sendToError(&sockets, &request->reqPeer, err));
}

/*
	perform the initial network startup
	this function is generally called only once at main stream startup
*/

sopair_t *netStart (VOID)
{
    INT psoc;
    sai_t lsa;
    struct servent *serv;
    static sopair_t sockets;

    if ((psoc = socket (AF_INET, SOCK_DGRAM, 0)) < 0) {
        LOG_MSG1 (LOG_ERR, "Cannot obtain primary socket: %m");
        return NULL;
    }

    lsa.sin_family      = AF_INET;
    lsa.sin_addr.s_addr = INADDR_ANY;

    if ((serv = getservbyname (TFTPSERVICE, "udp")) != NULL)
        lsa.sin_port = serv->s_port;
    else {
        LOG_MSG1 (LOG_WARNING, "service %s not found, using port %d",
                  TFTPSERVICE, TFTPPORT);
        lsa.sin_port = htons(TFTPPORT);
    }

    if (bind (psoc, (sa_t *) &lsa, sizeof(lsa)) < 0) {
        LOG_MSG1 (LOG_ERR, "Cannot bind to primary socket: %m");
        close (psoc);
        return NULL;
    }

    sockets.send = sockets.recv = psoc;

    return &sockets;
}

/*
	wait for incoming data on the current socket
	decode the initial message, check for legal requests
	and fill in the 'request' structure appropriately
*/

INT getRequest (sopair_t *sockets, tftpR_t *request, INT maxwait)
{
    tftpD_t recvBuffer;			/*	receive Buffer		*/
    INT recvLength;			/* received data in bytes	*/
    INT nameLength;			/* length of file name		*/
    INT rsaLength;			/* length of sin_addr_in struct	*/
    INT rtc;				/* general return code		*/ 
    CHAR *requestMode;			/* pointer to request mode text	*/
    struct timeval recvTimeout;		/* timeout for select		*/
    struct timeval *timeout;		/* needed for non-timed select	*/
    fd_set fdSel;			/* select bitmask		*/
    time_t startTime;			/* guess what !			*/

    CHAR *unsuppText;
    static CHAR unsuppFmt[] = "Unsupported transfer mode";

    startTime = time(NULL);

    while (1) {

        if (maxwait) {
            recvTimeout.tv_sec  = MAX(0, maxwait-(INT)(time(NULL)-startTime));
            recvTimeout.tv_usec = 0;
            timeout = &recvTimeout;
            LOG_MSG5 (LOG_DEBUG, "Waiting for x-fer request, timeout %d s",
                      recvTimeout.tv_sec);
        }
        else
            timeout = NULL;

        FD_ZERO(&fdSel);
		FD_SET(sockets->recv, &fdSel);

        rtc = select (sockets->recv + 1, &fdSel, NULL, NULL, timeout);

        if (rtc < 0) {
            if (errno == EINTR) {
                if(checkSignal())
                    return -1;
                else
                    continue;
            }
            LOG_MSG1 (LOG_ERR, "Error during select %d (read/GRQ): %m",
                      sockets->recv);
            return -1;
        }

        if (rtc == 0) 		/* timeout */
            return 0;

        rsaLength  = sizeof(request->reqPeer);
        recvLength = recvfrom (sockets->recv, &recvBuffer, sizeof(recvBuffer),
                               0, (sa_t *)&request->reqPeer, &rsaLength);

        if (recvLength < 0)  {
            if (errno == EINTR) {
                if(checkSignal())
                    return -1;
                else
                    continue;
            }
            LOG_MSG1 (LOG_ERR, "Receive error: %m");
            return -1;
        }

        if (recvLength < (INT)sizeof(tftpH_t)) {
            LOG_MSG1 (LOG_ERR, "Short request -- discarded");
            continue;
        }

        request->reqType = ntohs(recvBuffer.head.opCode);

        if (request->reqType != RT_RRQ &&
            request->reqType != RT_WRQ ) {
            if (request->reqType == RT_ACK) {
                LOG_MSG4 (LOG_DEBUG, "Received ACK on ERR/abort");
            }
            else {
                LOG_MSG1 (LOG_ERR, "Illegal transfer initialization [%d] by %s",
                          request->reqType, 
                          xinet_ntoa((in_addr_t)request->reqPeer.sin_addr.s_addr));
                sendToError(sockets, &request->reqPeer, PE_BADOP);
            }
            continue;
        }

        nameLength = strlen(recvBuffer.head.reqText);

        if (!nameLength) {
            LOG_MSG1 (LOG_ERR, "No file name in request packet");
            sendToError(sockets, &request->reqPeer, PE_BADOP);
            continue;
        }

        requestMode = recvBuffer.head.reqText + nameLength + 1;

        if (!strcmp(RMS_OCTET, requestMode))
            request->reqMode = RM_OCTET;
        else if (!strcmp(RMS_NETASCII, requestMode))
            request->reqMode = RM_NETASCII;
        else {
            unsuppText = (CHAR *)alloca(strlen(unsuppFmt)+
                                        strlen(requestMode)+8);
            if (unsuppText == NULL) {
                errNoDynmem(NULL);
                unsuppText = unsuppFmt;
            }
            else {
                sprintf (unsuppText, "%s '%s'", unsuppFmt, requestMode);
            }
            LOG_MSG1 (LOG_ERR, unsuppText);
            sendToError(sockets, &request->reqPeer, PE_UNDEF, unsuppText);
            continue;
        }

        if ((request->reqFile = (char *)malloc(nameLength+1)) == NULL) {
            errNoDynmem(NULL);
            sendToError(sockets, &request->reqPeer, PE_NOSPACE);
            return -1;
        }

        strcpy (request->reqFile, recvBuffer.head.reqText);

        break;
    }

    LOG_MSG4 (LOG_DEBUG, "%s request (%s) from %s for %s",
              request->reqType == RT_RRQ ? "Read" : "Write", requestMode,
              xinet_ntoa((in_addr_t)request->reqPeer.sin_addr.s_addr),
              request->reqFile);

    return 1;
}

/*
	set up a new connection to the current peer
	this function is called after a successful 'getRequest()',
	when a new child process has been created
*/

INT connectPeer (tftpR_t *request)
{
    INT newSocket;
    sai_t lsa;

    if ((newSocket = socket (AF_INET, SOCK_DGRAM, 0)) < 0) {
        LOG_MSG1 (LOG_ERR, "Cannot obtain slave socket: %m");
        return -1;
    }

    lsa.sin_family      = AF_INET;
    lsa.sin_addr.s_addr = INADDR_ANY;
    lsa.sin_port        = 0;

    if (bind (newSocket, (sa_t *) &lsa, sizeof(lsa)) < 0) {
        LOG_MSG1 (LOG_ERR, "Cannot bind to slave socket: %m");
        close (newSocket);
        return -1;
    }

    if (connect (newSocket,(sa_t *)&request->reqPeer,sizeof(sai_t)) < 0) {
        LOG_MSG1 (LOG_ERR, "Cannot connect via slave socket: %m");
        close (newSocket);
        return -1;
    }

    request->Rsock = request->Ssock = newSocket;
    request->xfrstat = request->iostat = 0;

    if (MsgLevel >= LVL_DEBUG) {
        sai_t localSock;
        INT sockLength = sizeof(localSock);
        if (getsockname(request->Rsock, (sa_t *)&localSock, &sockLength))
            localSock.sin_port = 0;
        LOG_MSG4 (LOG_DEBUG, "Connect to %s:%d "
                  "via port %d successful",
                  xinet_ntoa((in_addr_t)request->reqPeer.sin_addr.s_addr),
                  (INT)ntohs(request->reqPeer.sin_port),
                  (INT)ntohs(localSock.sin_port));
    }

    return 0;
}

/*
	receive a file from the peer
	we assume an already opened file and an active socket
        mode may be one of FM_OCTET or FM_NETASCII
*/

INT recvFile (tftpR_t *request)
{
    tftpD_t recvBuffer;		/* receive buffer (one packet)	*/
    tftpD_t ackBuffer;		/* acknowledge packet		*/
    INT recvLength;		/* received bytes		*/
    INT hsock;			/* highest socket+1 (select)	*/
    USHORT curBlock = 1;	/* netxt block awaited		*/
    USHORT lastBlock = 0;	/* last successful rec. block	*/
    INT retries = 0;		/* counter for retries		*/
    INT retrans = 0;		/* retransmission counter	*/
    INT done = 0;		/* flag for end-of-transfer	*/
    INT rtc;			/* general function return code	*/
    fd_set fdSel;		/* select bitmap		*/

    enum send_states { ST_RECVWAIT, ST_RECV, ST_PUTDATA,
                       ST_ACKWAIT, ST_ACK } state;

    static char *state_names[] = { "ST_RECVWAIT", "ST_RECV", "ST_PUTDATA", 
                                   "ST_ACKWAIT","ST_ACK"};

    struct timeval timeout;	/* select timeout		*/
 
    static struct timeval recvTimeout = { RECV_TIMEOUT, 0 },
                          ackTimeout  = { RACK_TIMEOUT, 0 };

    hsock = MAX(request->Rsock, request->Ssock) + 1;

/*
	Loop through the sate-machine that implements the send algorithm
*/

    for (state = ST_ACKWAIT; ; ) {

        LOG_MSG5 (LOG_DEBUG,"recvFile: state %d: %s",
                  state, state_names[state]);

        switch (state) {
/*
	wait for data packet from peer
	this is actually a 'wait for any reply',
	because we shoud be prepared to receive an packet,
	especially error packets.

	On timeout, we check the retry-count.
	This counter will be reset on any sucessfully transferred packet.
	If retries are still allowed, we re-ack the latest packet.
*/
            case ST_RECVWAIT:
                FD_ZERO(&fdSel);
                FD_SET(request->Rsock, &fdSel);
                timeout = recvTimeout;

                rtc = select (hsock, &fdSel, NULL, NULL, &timeout);
                if (rtc < 0) {
                    LOG_MSG1 (LOG_ERR, "Error during select (read/RCV): %m");
                    return -1;
                }
                else if (rtc == 0) {		/* timeout */
                    if (retries >= MAXRETRIES) {
                        LOG_MSG1 (LOG_ERR,
                                  "Too many retries while waiting for data packet");
                        return -1;
                    }
                    else {
                        retries++;
                        state = ST_ACKWAIT;
                        break;
                    }
                }
                state = ST_RECV;
                break;
/*
	The received packed should be at least of length sizeof(tftpD_t)
	(usually 4 bytes). Short packets are silently discarded:
	we proceed as with timeouts above.

	Now let's check the opcode of the received packets.
	All non-data packets are considered fatal transfer errors, which
	lead to an abort of the current transfer.
*/
            case ST_RECV:
                recvLength = recv (request->Rsock, 
                                   &recvBuffer, sizeof(tftpD_t), 0);
                if (recvLength < 0) {
                    LOG_MSG1 (LOG_ERR, "Error during receive: %m");
                    return -1;
                }

                if (recvLength < (INT)sizeof(tftpH_t)) {
                    LOG_MSG1 (LOG_ERR, "Short DATA packet");
                    if (retries >= MAXRETRIES) {
                        LOG_MSG1 (LOG_ERR, 
                                  "Too many retries while waiting for ACK");
                        return -1;
                    }
                    else {
                        retries++;
                        state = ST_ACKWAIT;
                        break;
                    }
                }

                if (ntohs(recvBuffer.head.opCode) == RT_DATA) {
/*
	We got the desired data packet, so lets check the block number.
	Only 2 cases are valid: (current-block) and (current-block -1),

	If we receive (current-block -1) we first check the retry count.
	If retries are still allowed, we simply acknowledge the packet
	and discard the data, because this is apparently a re-sent packet.
*/
                    if (ntohs(recvBuffer.head.blockNo) == lastBlock) {
                        if (retrans > MAXRETRANS) {
                            LOG_MSG1 (LOG_ERR, "Too many retransmissions");
                            return -1;
                        }
                        retrans++;
                        state = ST_ACKWAIT;
                    }
                    else if (ntohs(recvBuffer.head.blockNo) != curBlock) {
                        LOG_MSG1 (LOG_WARNING, 
                                  "Received out of sequence DATA no %d"
                                  " while awaiting %d",
                                  ntohs(recvBuffer.head.blockNo), lastBlock);
                        if (retries >= MAXRETRIES) {
                            LOG_MSG1 (LOG_ERR, 
                                      "Too many retries while waiting for ACK");
                            return -1;
                        }
                        else {
                            retries++;
                            state = ST_ACKWAIT;
                        }
                    }
                    state = ST_PUTDATA;
                    break;
                }
/*
	Non-DATA packet: terminate transfer anyway
*/
                else {
                    LOG_MSG1 (LOG_ERR, 
                              "Unexpected packet type [%d] "
                              "while waiting for DATA", 
                              (INT)ntohs(recvBuffer.head.opCode));        
                    return -1;
                }
                break;
/*
	Now, this is what we expect to come in next.
	Let's look at the length: anything  < SEGSIZE is a 'final packet'.
	in this case prepare for closing the transfer.
*/
            case ST_PUTDATA:
                recvLength -= sizeof(tftpH_t);
                done = (recvLength < SEGSIZE);
                request->xfrstat += (count_t)recvLength;
                if (recvLength) {
                    if (putXferBuffer (request->reqFd, recvBuffer.data, 
                                       recvLength, request->reqMode, done,
                                       &request->iostat) != recvLength) {
                        LOG_MSG1 (LOG_ERR, "Cannot write local data: %m");
                        sendError(request, PE_NOSPACE);
                        return -1;
                    }
                }
                LOG_MSG5 (LOG_DEBUG,"recvFile: %d bytes to file", recvLength);
                state = ST_ACKWAIT;
                lastBlock = curBlock++;
                break;
/*
	wait for send channel to become ready
	we don't tolerate timeouts at this point
	on success, proceed with 'send'
*/
            case ST_ACKWAIT:
                FD_ZERO(&fdSel);
                FD_SET(request->Ssock, &fdSel);
                timeout = ackTimeout;

                rtc = select (hsock, NULL, &fdSel, NULL, &timeout);
                if (rtc < 0) {
                    LOG_MSG1 (LOG_ERR, "Error during select (write/ACK): %m");
                    return -1;
                }
                else if (rtc == 0) {		/* timeout */
                    LOG_MSG1 (LOG_ERR, "Timeout during select (write/ACK)");
                    return -1;
                }
                state = ST_ACK;
                break;
/*
	send an ACK for the block
*/
            case ST_ACK:
                ackBuffer.head.opCode  = htons(RT_ACK);
                ackBuffer.head.blockNo = htons(lastBlock);

                if (send (request->Ssock, &ackBuffer, sizeof(ackBuffer), 0) != 
                    sizeof(ackBuffer)) {
                    LOG_MSG1 (LOG_ERR, "Error during send: %m"); 
                    return -1;
                }
                LOG_MSG5 (LOG_DEBUG,"recvFile: ACK for block %d%s",
                          lastBlock, lastBlock ? "": " (init transfer)");
                if (done) {
                    
                    LOG_MSG3 (LOG_INFO,"WRQ successful");
                    return 0;
                }

                state = ST_RECVWAIT;
                retries = retrans = 0;
                break;
        }
    }
    return -1;
}

/*
	send a file to the peer
	we assume an already opened file and an active socket
        mode may be one of FM_OCTET or FM_NETASCII
*/

INT sendFile (tftpR_t *request)
{
    tftpD_t sendBuffer, ackBuffer; 
    INT bufLength, recvLength, sendLength, hsock;
    unsigned short curBlock = 0 , lastBlock = 0;
    INT retries = 0, retrans = 0, done = 0, rtc;
    fd_set fdSel;
    enum send_states {ST_GETDATA, ST_SENDWAIT, ST_SEND,
                      ST_ACKWAIT, ST_ACK } state;
    static char *state_names[]  = { "ST_GETDATA", "ST_SENDWAIT", "ST_SEND",
                                    "ST_ACKWAIT", "ST_ACK" };
    struct timeval timeout; 
    static struct timeval sendTimeout = { SEND_TIMEOUT, 0 },
                          ackTimeout  = { SACK_TIMEOUT, 0 };

    hsock = MAX(request->Rsock, request->Ssock) + 1;
/*
	Loop through the sate-machine that implements the send algorithm
*/
    for (state = ST_GETDATA; ; ) {

        LOG_MSG5 (LOG_DEBUG,"sendFile: state %d: %s",
                  state, state_names[state]);

        switch (state) {
/*
	get the next chunk of data from disk
	on success fall through to 'send wait'
*/ 
            case ST_GETDATA:
                bufLength = getXferBuffer (request->reqFd, sendBuffer.data, 
                                           SEGSIZE, request->reqMode,
                                           &request->iostat);
                if (bufLength < 0) {
                    LOG_MSG1 (LOG_ERR, "Cannot read local data: %m");
                    return -1;
                }
                else if (bufLength < SEGSIZE)
                    done = 1;
                lastBlock = curBlock++;
                retries = retrans = 0;
                LOG_MSG5 (LOG_DEBUG,"sendFile: %d bytes from file", bufLength);
                state = ST_SENDWAIT;
                break;
/*
	wait for send channel to become ready
	we don't tolerate timeouts at this point
	on success, proceed with 'send'
*/
            case ST_SENDWAIT:
                FD_ZERO(&fdSel);
                FD_SET(request->Ssock, &fdSel);
                timeout = sendTimeout;

                rtc = select (hsock, NULL, &fdSel, NULL, &timeout);
                if (rtc < 0) {
                    LOG_MSG1 (LOG_ERR, "Error during select (write/SND): %m");
                    return -1;
                }
                else if (rtc == 0) {		/* timeout */
                    LOG_MSG1 (LOG_ERR, "Timeout during select (write/SND)");
                    return -1;
                }
                state = ST_SEND;
                break;
/*
	send the current buffer
	and fall throug 'wait for ack'
*/
            case ST_SEND:
                sendBuffer.head.opCode  = htons(RT_DATA);
                sendBuffer.head.blockNo = htons(curBlock);
                sendLength = sizeof(tftpH_t) + bufLength;

                if (send (request->Ssock, &sendBuffer, sendLength, 0) != sendLength) {
                    LOG_MSG1 (LOG_ERR, "Error during send: %m"); 
                    return -1;
                }
                LOG_MSG5 (LOG_DEBUG,"sendFile: block %d of %d bytes sent",
                          curBlock, sendLength); 
                request->xfrstat += (count_t)sendLength;
                state = ST_ACKWAIT;
                break;
/*
	wait for ACK from peer
	this is actually a 'wait for any reply',
	because we shoud be prepared to receive an packet,
	especially error packets.

	On timeout, we check the retry-count.
	This counter will be reset on any sucessfully transferred packet.
	If retries are still allowesd, we re-send the current packet.
*/
            case ST_ACKWAIT:
                FD_ZERO(&fdSel);
                FD_SET(request->Rsock, &fdSel);
                timeout = ackTimeout;

                rtc = select (hsock, &fdSel, NULL, NULL, &timeout);
                if (rtc < 0) {
                    LOG_MSG1 (LOG_ERR, "Error during select (read/ACK): %m");
                    return -1;
                }
                else if (rtc == 0) {		/* timeout */
                    if (retries >= MAXRETRIES) {
                        LOG_MSG1 (LOG_ERR,
                                  "Too many retries while waiting for ACK");
                        return -1;
                    }
                    else {
                        retries++;
                        state = ST_SENDWAIT;
                        break;
                    }
                }
                state = ST_ACK;
                break;
/*
	The received packed should be at least of length sizeof(tftpD_t)
	(usually 4 bytes). Short packets are silently discarded:
	we proceed as with timeouts above.

	Now let's check the opcode of the received packets.
	All non-ACK packets are considered fatal transfer errors, which
	lead to an abort of the current transfer.
*/
            case ST_ACK:
                recvLength = sizeof(tftpH_t) + SEGSIZE;
                recvLength = recv (request->Rsock, &ackBuffer, recvLength, 0);
                if (recvLength < 0) {
                    LOG_MSG1 (LOG_ERR, "Error during recv: %m");
                    return -1;
                }

                if (recvLength < (INT)sizeof(tftpH_t)) {
                    LOG_MSG1 (LOG_ERR, "Short ACK packet");
                    if (retries >= MAXRETRIES) {
                        LOG_MSG1 (LOG_ERR, 
                                  "Too many retries while waiting for ACK");
                        return -1;
                    }
                    else {
                        retries++;
                        state = ST_SENDWAIT;
                        break;
                    }
                }

                switch (ntohs(ackBuffer.head.opCode)) {
/*
	We got the desired ACK packet, so lets check the block number.
	Only 2 cases are valid: (current-block) and (current-block -1),
	because we only send a new block, if the last one was ACKed.

	If the current block was acked, proceed with the next one, or,
	if this was the last block, quit.

	If the previous blcok was acked, check the retry-count, and
	re-send if possible.
*/
                    case RT_ACK:
                        LOG_MSG5 (LOG_DEBUG,"sendFile: ACK for block %d",
                                  (INT)ntohs(ackBuffer.head.blockNo));
                        if (ntohs(ackBuffer.head.blockNo) == curBlock) {
                            if (done) {
                                LOG_MSG3 (LOG_INFO,
                                          "RRQ successful");
                                return 0;
                            }
                            state = ST_GETDATA;
                        }
                        else if (ntohs(ackBuffer.head.blockNo) == lastBlock) {
                            if (retrans > MAXRETRANS) {
                                LOG_MSG1 (LOG_ERR, "Too many retransmissions");
                                return -1;
                            }
                            retrans++;
                            state = ST_SENDWAIT;
                        }
                        else {
                            LOG_MSG1 (LOG_ERR, 
                                      "Received out of sequence ACK no %d"
                                      " while awaiting %d",
                                      ntohs(ackBuffer.head.blockNo),
                                      lastBlock);
                            return -1;
                        }
                        break;
/*
	Non-ACK packet: terminate transfer anyway
	Error packets give us a chance to obtain more info about the failure
*/
                    case RT_ERROR:
                        LOG_MSG1 (LOG_ERR, "Peer reports serror [%d]: %s",
                                  (INT)ntohs(ackBuffer.head.errCode), 
                                  ackBuffer.data);
                        return -1;
                        break;

                    default:
                        LOG_MSG1 (LOG_ERR, 
                                  "Unknown type [%d] packet while awaiting ACK",
                                  (INT)ntohs(ackBuffer.head.opCode));        
                        return -1;
                        break;
                }
                break;
        } 
    }
    return -1;
}
